#include <ddraw.h>
#include <sys/timeb.h>
#include <stdio.h>
#include <conio.h>
#include <math.h>
#define PI 3.1415926535897932384626433832795028841971693993751
#include "3d.h"

LPDIRECTDRAW7			m_lpDD;
LPDIRECTDRAWSURFACE7	m_lpDDSFront;
LPDIRECTDRAWSURFACE7	m_lpDDSBack;


HRESULT CreateSurface( DWORD size_x, DWORD size_y, LPDIRECTDRAWSURFACE7 *surface ) {
	HRESULT				dwResult;
	DDSURFACEDESC2		ddsd;

	ZeroMemory( &ddsd, sizeof( ddsd ) );
	ddsd.dwSize			= sizeof( ddsd );
	ddsd.dwFlags		= DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH;
	ddsd.dwWidth		= size_x;
	ddsd.dwHeight		= size_y;
	ddsd.ddsCaps.dwCaps	= DDSCAPS_OFFSCREENPLAIN;

	dwResult			= m_lpDD->CreateSurface( &ddsd, surface, NULL );
	return dwResult;
}

HRESULT InitDDraw( HWND hWnd, DWORD res_x, DWORD res_y, DWORD res_bit ) {
	m_lpDD			= NULL;
	m_lpDDSFront	= NULL;
	m_lpDDSBack		= NULL;

    HRESULT			dwResult;
	DDSURFACEDESC2	ddsd;
	DDSCAPS2		ddscaps;

    dwResult = DirectDrawCreateEx( NULL, ( void** )&m_lpDD, IID_IDirectDraw7, NULL );
    if ( dwResult != DD_OK ) {
        return dwResult;
    }

    dwResult = m_lpDD->SetCooperativeLevel( hWnd, DDSCL_EXCLUSIVE | 
													DDSCL_FULLSCREEN | 
													DDSCL_ALLOWREBOOT );
    if ( dwResult != DD_OK ) {
        return dwResult;
    }

    dwResult = m_lpDD->SetDisplayMode( res_x, res_y, res_bit, 0, 0 );
    if( dwResult != DD_OK ) {
        return dwResult;
    }

	ZeroMemory( &ddsd, sizeof( ddsd ) );
	ddsd.dwSize				= sizeof( ddsd );
	ddsd.dwFlags			=	DDSD_CAPS | 
								DDSD_BACKBUFFERCOUNT;
	ddsd.ddsCaps.dwCaps		=	DDSCAPS_PRIMARYSURFACE | 
								DDSCAPS_FLIP | 
								DDSCAPS_COMPLEX;
	ddsd.dwBackBufferCount	= 1	;

	dwResult = m_lpDD->CreateSurface( &ddsd, &m_lpDDSFront, NULL );
	if( dwResult != DD_OK ) {
		return dwResult;
	} 

    ZeroMemory( &ddscaps, sizeof( ddscaps ) );
    ddscaps.dwCaps = DDSCAPS_BACKBUFFER;

	dwResult = m_lpDDSFront->GetAttachedSurface( &ddscaps, &m_lpDDSBack );
    if( dwResult != DD_OK ) {
		return dwResult;
	}

	return dwResult;
}


void CleanDDraw() {
	if ( m_lpDD ) {
		if( m_lpDDSBack )	m_lpDDSBack->Release();
		if( m_lpDDSFront )	m_lpDDSFront->Release();		
		m_lpDDSFront	= NULL;
		m_lpDDSBack		= NULL;
		m_lpDD->RestoreDisplayMode();
		m_lpDD->Release();
		m_lpDD			= NULL;
	}
}

LRESULT WINAPI MsgProc( HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam ) {
    switch( msg ) {
        case WM_DESTROY:
            PostQuitMessage( 0 );
            return 0;

        case WM_PAINT:
            ValidateRect( hWnd, NULL );
            return 0;
    }

    return DefWindowProc( hWnd, msg, wParam, lParam );
}

void InitWin( WNDCLASSEX *wc, HWND *hWnd, char *class_name, char *win_name, int x1, int y1, int x2, int y2 ) {
    wc->cbSize			= sizeof(WNDCLASSEX);
	wc->style			= CS_CLASSDC;
	wc->lpfnWndProc		= MsgProc;
	wc->cbClsExtra		= 0L;
	wc->cbWndExtra		= 0L;
	wc->hInstance		= GetModuleHandle(NULL);
	wc->hIcon			= NULL;
	wc->hCursor			= NULL;
	wc->hbrBackground	= NULL;
	wc->lpszMenuName	= NULL;
	wc->lpszClassName	= class_name;
	wc->hIconSm			= NULL;

    RegisterClassEx(wc);

    ( *hWnd ) = CreateWindowEx( WS_EX_TOPMOST, class_name, win_name, 
                              WS_OVERLAPPEDWINDOW, x1, y1, x2, y2,
                              GetDesktopWindow(), NULL, wc->hInstance, NULL );
}

void CleanWin( WNDCLASSEX *wc, char *class_name ) {
	UnregisterClass( class_name, wc->hInstance );
}

/***************************************************************************/
#define RES_X 800
#define RES_Y 600
#define RES_B 32
#define CLASS_N "Direct Draw Test"
#define WIN_N "Simulace"
#define POC 17225
//#define POC 50000
#define BS 100000
#define PREC 0.000005
//#define PREC 0.1
#define OKOLI 500000

typedef int snt;
typedef long sis;

struct	_timeb start, stop;
int		i = 0;
double	fps = 0;
char	sfps[20];
HDC		hdc;

bod		B = {-1000000, 0, 0};
vektor	up = {-1, 0, 0}, vp = {0, 0, 1};
tip		d = 1000;
tip alpha = -0.5;

snt bn;

struct _box {
	snt sod, sdo;
	bod lr;
	int rp;
} *box;

bod star[POC];
obod sobr[POC];
bod *boxx;
obod *bobr;
char color[POC];

char r, g, b;

DDSURFACEDESC2 dds;
DDBLTFX dblx;

RECT rcdes = {0, 0, RES_X, RES_Y};
unsigned char *pix;
HRESULT hr;

#define EOL 13

void readln(FILE *vstup, char *s, int poc) {
  int i = 0, c = 0;
  while ( ((c = fgetc(vstup)) != EOL) & (i < poc)) {
    s[i] = (char)c;
    i ++;
  }
  s[i] = '\n';
  s[i + 1] = 0;
}

void readln(FILE *vstup) {
  while ( fgetc(vstup) != EOL ) {
  }
}

int nacti(bod *co, char *color, char *soubor, int pocet, int krok) {
  FILE *vstup;

  char s[256];
  int c = 0, i, j, k;
  char b = 8;

  float x, y, z;
  int h;

//  printf("Nacitam ze souboru: %s\n\n\n", soubor);

  vstup = fopen(soubor, "rb");
  if (vstup == NULL) {
//    printf("Chyba pri otevirani souboru %s", soubor);
    return 0;
  }

  if (pocet <= 0) {
//    printf("Spatne zadana velikost pole");
    return 0;
  }

  while ( (c = fgetc(vstup)) == '#') {
    readln(vstup, s, 256);
    printf("%s", s);
  }

  i = 0;
  j = 0;
  k = 0;

  while (feof(vstup) == 0) {
    c = fgetc(vstup);
    while ( (c == '#') | (c == 10) | (c == EOL) | (c == ' ') ){
      if (c == '#') {
	readln(vstup);
	printf(".");
	b ++;
      }
      c = fgetc(vstup);
    }
    ungetc(c, vstup);

    fscanf(vstup, "%f", &x);
    fscanf(vstup, "%f", &y);
    fscanf(vstup, "%f", &z);
    fscanf(vstup, "%d", &h);

    if (j >= krok) {
      if (i < pocet) {
	co[i].x = x*100000;
	co[i].y = y*100000;
	co[i].z = z*100000;
//	color[i] = b;
	color[i] = h;
      }
      i ++;
      j = 0;
    }
    j ++;
    k ++;
  }

  if (i >= pocet) {
    co[pocet - 1].x = 0;
    co[pocet - 1].y = 0;
    co[pocet - 1].z = 0;
//  co[pocet - 1].h = 0;
  }

  fclose(vstup);
  return k;
}

void sets() {
    nacti(star, color, "galaxy.dat", POC, 1);
/*
	snt j;
	tip t, r;
	for (j = 0; j < POC; j ++) {
		t = (tip)rand()/0x7fff*2*PI;
		r = (tip)rand()+100;
		star[j].x = cos(t)*r*0.3;
		star[j].y = sin(t)*r*0.3;
		star[j].z = 1/r*10000000;
		star[j].z += 10;

//		star[j].x = (tip)rand()*0.2;
//		star[j].y = (tip)rand()*0.2;
//		star[j].z = (tip)rand()*0.2;

	}
*/
}

inline void bod32( unsigned char *pix, LONG l, int x, int y, char r, char g, char b ) {
	if ((x >= 0) && (x < RES_X) && (y >= 0) && (y < RES_Y)) {
		pix[4*x + y*l] = b;
		pix[4*x+1 + y*l] = g;
		pix[4*x+2 + y*l] = r;
		pix[4*x+3 + y*l] = 0;
	}
}

inline bool comp(snt i, snt j) {
	return 
	   (( ((sis)(star[i].x/BS)) == ((sis)(star[j].x/BS)) )&&
		( ((sis)(star[i].y/BS)) == ((sis)(star[j].y/BS)) )&&
		( ((sis)(star[i].z/BS)) == ((sis)(star[j].z/BS)) ));
}

void trid() {
	snt j, k, i;
	bod pom;
	for (i = 0; i < POC; i++) {
		for (j = i; j < POC; j ++) {
			if (!comp(i, j)) break;
		}
		for (k = j + 1; k < POC; k ++) {
			if (comp(i, k)) {
				pom = star[j];
				star[j] = star[k];
				star[k] = pom;
				j ++;
			}
		}
		i = j;
	}
}

void boxnum() {
	bn = 0;
	snt i, j = 0;
	for (i = 0; i < POC; i ++) {
		if (!comp(i, j)) {
			bn ++;
			j = i;
		}
	}
}

void binit() {
	snt i, j = 0, k = 0;
	for (i = 0; i < POC; i ++) {
		if (!comp(i, j)) {
			box[k].sod = j;
			box[k].sdo = i;
			boxx[k].x = BS*(sis)(star[j].x/BS)+BS/2;
			boxx[k].y = BS*(sis)(star[j].y/BS)+BS/2;
			boxx[k].z = BS*(sis)(star[j].z/BS)+BS/2;
			j = i;
			k ++;
		}
	}
}


void FPS() {
	if (m_lpDDSBack->GetDC(&hdc) == DD_OK) {
		sprintf(sfps, "FPS: %.1f", fps);
		SetBkColor( hdc, RGB( 0, 0, 0 ) );
		SetTextColor( hdc, RGB( 0, 0, 255 ) );
		TextOut( hdc, 0, 0, sfps, lstrlen(sfps) );
		m_lpDDSBack->ReleaseDC(hdc);
	}
	i ++;
	if (i > 50)
	{
		_ftime(&stop);
		fps = i/((stop.time - start.time) + 0.001*(stop.millitm - start.millitm));
		_ftime(&start);
		i = 0;
	}
}

void renderall(void) {
	snt i, j;
	prumet(star, sobr, B, up, vp, d, 0, POC);
	prumet(boxx, bobr, B, up, vp, d, 0, bn);
	for (i = 0; i < bn; i ++) {
		box[i].lr = B;
 		for (j = box[i].sod; j < box[i].sdo; j ++) {
			sobr[j].x -= bobr[i].x;
			sobr[j].y -= bobr[i].y;
//			sobr[j].z = bobr[i].z;
		}
	}
}

void move() {
	B.x += up.x*500;
	B.y += up.y*500;
	alpha += 0.0005;
	up.x = cos(alpha);
	up.y = sin(alpha);
}

void InitOther() {
	_ftime(&start);
	i = 0;
	fps = 0;

	ZeroMemory(&dds, sizeof(dds));
	dds.dwSize = sizeof(dds);
	ZeroMemory(&dblx, sizeof(dblx));
	dblx.dwSize = sizeof(dblx);
	dblx.dwFillColor = RGB(0, 0, 0);

	sets();
	trid();
	boxnum();
	box = new _box[bn];
	boxx = new bod[bn];
	bobr = new obod[bn];
	binit();
	move();
	renderall();
}

void priority() {
	snt i;
	tip s, vzd;

	for (i = 0; i < bn; i ++) {
		vzd = ((B.x-boxx[i].x)*(B.x-boxx[i].x) + (B.y-boxx[i].y)*(B.y-boxx[i].y) + (B.z-boxx[i].z)*(B.z-boxx[i].z));
		s = ((B.x-box[i].lr.x)*(B.x-box[i].lr.x) + (B.y-box[i].lr.y)*(B.y-box[i].lr.y) + (B.z-box[i].lr.z)*(B.z-box[i].lr.z));
		if (s/vzd < PREC) {
			box[i].rp = 0;
		} else {
			box[i].rp = 1;
		}
		if ( (abs(B.x-boxx[i].x) < OKOLI) && (abs(B.y-boxx[i].y) < OKOLI) && (abs(B.z-boxx[i].z) < OKOLI)) box[i].rp = 1;
	}
}

void renderprior() {
	snt i, j;
	prumet(boxx, bobr, B, up, vp, d, 0, bn);
	for (i = 0; i < bn; i ++) {
		if (box[i].rp == 1) {
			prumet(star, sobr, B, up, vp, d, box[i].sod, box[i].sdo);
			box[i].lr = B;
 			for (j = box[i].sod; j < box[i].sdo; j ++) {
				sobr[j].x -= bobr[i].x;
				sobr[j].y -= bobr[i].y;
//				sobr[j].z = bobr[i].z;
			}
		}
	}
}

void CleanOther() {
}

void Render() {

	priority();
	renderprior();
//	prumet(star, sobr, B, up, vp, d, 0, POC);
	move();

	m_lpDDSBack->Blt(&rcdes, NULL, NULL, DDBLT_WAIT|DDBLT_COLORFILL , &dblx);
	hr = m_lpDDSBack->Lock(NULL, &dds, DDLOCK_WAIT, NULL);
	if (hr == DD_OK) {
		pix = (unsigned char *)(dds.lpSurface);
		snt i, j;
		for (j = 0; j < bn; j ++) {
			for (i = box[j].sod; i < box[j].sdo; i ++) {
				if ((sobr[i].z > 0)) {
//					r = g = b = 255 - (7 + color[i])*100;
					r = g = b = 255;
					bod32(pix, dds.lPitch, RES_X/2+sobr[i].x+bobr[j].x, RES_Y/2+sobr[i].y+bobr[j].y, r-sobr[i].z*0.0001, g-sobr[i].z*0.0001, b-sobr[i].z*0.0001);
//					bod32(pix, dds.lPitch, RES_X/2+sobr[i].x+bobr[j].x, RES_Y/2+sobr[i].y+bobr[j].y, r, g, b);
				}
			}

		}
/*		
		for (i = 0; i < POC; i ++) {
				if ((sobr[i].z > 0)) {
//					r = g = b = 255 - (7 + color[i])*100;
					r = g = b = 255;
					bod32(pix, dds.lPitch, RES_X/2+sobr[i].x, RES_Y/2+sobr[i].y, r-sobr[i].z*0.001, g-sobr[i].z*0.001, b-sobr[i].z*0.001);
				}
		}
*/
		m_lpDDSBack->Unlock(NULL); 
	}


	FPS();

	m_lpDDSFront->Flip( NULL, DDFLIP_WAIT );
}


INT WINAPI WinMain( HINSTANCE hInst, HINSTANCE, LPSTR, INT ) {
	WNDCLASSEX	wc;
	HWND		hWnd;
	BOOL		gm = TRUE;
	MSG			msg;

	InitWin( &wc, &hWnd, CLASS_N, WIN_N, 100, 100, 200, 200 );

	if( SUCCEEDED( InitDDraw( hWnd, RES_X, RES_Y, RES_B ) ) ) {
		InitOther();
        ShowWindow( hWnd, SW_SHOWDEFAULT );
        UpdateWindow( hWnd );

		while ( gm ) {
			if( PeekMessage( &msg, NULL, 0, 0, PM_NOREMOVE ) ) {
				if( gm = GetMessage( &msg, NULL, 0, 0 ) ) {
					TranslateMessage( &msg );
					DispatchMessage( &msg );
				}
			}
			else Render();
		}
    }

    CleanOther();
    CleanDDraw();
    CleanWin(&wc, CLASS_N);

    return 0;
}
